import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions.normal import Normal
from .utils import init

# ---- Categorical ------------------------------------------------------------

class Categorical(nn.Module):
    """
    Categorical action distribution.

    Parameters
    ----------
    num_inputs : int
        Size of input feature maps.
    num_outputs : int
        Number of action in action space.

    Attributes
    ----------
    linear: nn.Module
        Maps the incoming feature maps to probabilities over actions in output space.
    """
    def __init__(self, num_inputs, num_outputs):
        super(Categorical, self).__init__()
        self.linear = nn.Linear(num_inputs, num_outputs)

    def forward(self, x, deterministic=False):
        """
        Predict distribution parameters from x (obs features) and return
        actions (sampled and clipped), sampled action log
        probability and distribution entropy.

        Parameters
        ----------
        x : torch.tensor
            Feature maps extracted from environment observations.
        deterministic : bool
            Whether to randomly sample action from predicted distribution or take the mode.

        Returns
        -------
        action: torch.tensor
            Predicted next action.
        clipped_action: torch.tensor
            Predicted next action (clipped to be within action space).
        logp_action : torch.tensor
            Log probability of `action` according to the predicted action distribution.
        entropy_dist : torch.tensor
            Entropy of the predicted action distribution.
        """

        # Predict distribution parameters
        x = self.linear(x)

        # Create distribution and sample action
        action_dist = torch.distributions.Categorical(logits=x)
        self.action_dist = action_dist # ugly hack to handle sac discrete case

        if deterministic:
            action = clipped_action = action_dist.probs.argmax(dim=-1, keepdim=True)
        else:
            action = clipped_action = action_dist.sample().unsqueeze(-1)

        # Action log probability
        logp_action = action_dist.log_prob(action.squeeze( -1)).unsqueeze(-1)

        # Distribution entropy
        entropy_dist = action_dist.entropy().mean()

        return action, clipped_action, logp_action, entropy_dist


    def evaluate_actions(self, x, action):
        """
        Return log prob of action under the distribution generated from
        x (obs features). Also return entropy of the generated distribution.

        Parameters
        ----------
        x : torch.tensor
            obs feature map obtained from a policy_net.
        action : torch.tensor
            Evaluated action.

        Returns
        -------
        logp_action : torch.tensor
            Log probability of `action` according to the action distribution
            predicted.
        entropy_dist : torch.tensor
            Entropy of the action distribution predicted.
        """

        # Predict distribution parameters
        x = self.linear(x)

        # Create distribution
        action_dist = torch.distributions.Categorical(logits=x)

        # Evaluate log prob of action under action_dist
        logp_action = action_dist.log_prob(action.squeeze(-1)).unsqueeze(-1).sum(-1, keepdim=True)

        # Distribution entropy
        entropy_dist = action_dist.entropy().mean()

        return logp_action, entropy_dist

# ---- Gaussian ---------------------------------------------------------------

class DiagGaussian(nn.Module):
    """
    Gaussian action distribution.

    Parameters
    ----------
    num_inputs : int
        Size of input feature maps.
    num_outputs : int
        Number of action in action space.
    predict_log_std : bool
        Whether to use a nn.linerar layer to predict the output actions std.

    Attributes
    ----------
    mean: nn.Module
        Maps the incoming feature maps to output action mean values.
    log_std : nn.Module or nn.Parameter
        If predict_log_std is True Maps the incoming feature maps to output
        action std values. Otherwise, the std values are a learnable nn.Parameter.
    LOG_STD_MAX : float
        Maximum std allowed value. Used for clipping.
    LOG_STD_MIN : float
        Minimum std allowed value. Used for clipping.
    """

    LOG_STD_MAX = 2
    LOG_STD_MIN = -20

    def __init__(self, num_inputs, num_outputs, predict_log_std=False):
        super(DiagGaussian, self).__init__()

        self.predict_log_std = predict_log_std

        init_ = lambda m: init(m, nn.init.orthogonal_, lambda x: nn.init.constant_(x, 0))
        self.mean = init_(nn.Linear(num_inputs, num_outputs))

        if predict_log_std:
            self.log_std = init_(nn.Linear(num_inputs, num_outputs))
        else:
            self.log_std = nn.Parameter(torch.zeros(num_outputs).unsqueeze(0))

    def forward(self, x, deterministic=False):
        """
        Predict distribution parameters from x (obs features) and return
        actions (sampled and clipped), sampled action log
        probability and distribution entropy.

        Parameters
        ----------
        x : torch.tensor
            Feature maps extracted from environment observations.
        deterministic : bool
            Whether to randomly sample action from predicted distribution or take the mode.

        Returns
        -------
        action: torch.tensor
            Predicted next action.
        clipped_action: torch.tensor
            Predicted next action (clipped to be within action space).
        logp_action : torch.tensor
            Log probability of `action` according to the predicted action distribution.
        entropy_dist : torch.tensor
            Entropy of the predicted action distribution.
        """
        # Predict distribution parameters
        action_mean = self.mean(x)
        action_logstd = self.log_std(x) if self.predict_log_std else\
            torch.zeros(action_mean.size()).to(x.device) + self.log_std

        # action_logstd = torch.clamp(action_logstd, self.LOG_STD_MIN, self.LOG_STD_MAX)

        # Create distribution and sample action
        action_dist = Normal(action_mean, action_logstd.exp())

        if deterministic:
            action = action_mean
        else:
            action = action_dist.sample()

        # Apply clipping to avoid being outside action space
        clipped_action = torch.clamp(action, -1, 1)

        # Action log probability
        logp_action = action_dist.log_prob(action).sum(-1, keepdim=True)

        # Distribution entropy
        entropy_dist = action_dist.entropy().sum(-1).mean()

        return action, clipped_action, logp_action, entropy_dist

    def evaluate_actions(self, x, action):
        """
        Return log prob of action under the distribution generated from
        x (obs features). Also return entropy of the generated distribution.

        Parameters
        ----------
        x : torch.tensor
            obs feature map obtained from a policy_net.
        action : torch.tensor
            Evaluated action.

        Returns
        -------
        logp_action : torch.tensor
            Log probability of `action` according to the action distribution
            predicted.
        entropy_dist : torch.tensor
            Entropy of the action distribution predicted.
        """

        # Predict distribution parameters

        action_mean = self.mean(x)
        action_logstd = self.log_std(x) if self.predict_log_std else\
            torch.zeros(action_mean.size()).to(x.device) + self.log_std

        # action_logstd = torch.clamp(action_logstd, self.LOG_STD_MIN, self.LOG_STD_MAX)

        # Create distribution
        action_dist = Normal(action_mean, action_logstd.exp())

        # Evaluate log prob of action under action_dist
        logp_action = action_dist.log_prob(action).sum(-1, keepdim=True)

        # Distribution entropy
        entropy_dist = action_dist.entropy().sum(-1).mean()

        return logp_action, entropy_dist

# ---- Squashed Gaussian ------------------------------------------------------

class SquashedGaussian(nn.Module):
    """
    Squashed Gaussian action distribution.

    Parameters
    ----------
    num_inputs : int
        Size of input feature maps.
    num_outputs : int
        Number of action in action space.
    predict_log_std : bool
        Whether to use a nn.linerar layer to predict the output actions std.

    Attributes
    ----------
    mean: nn.Module
        Maps the incoming feature maps to output action mean values.
    log_std : nn.Module or nn.Parameter
        If predict_log_std is True Maps the incoming feature maps to output
        action std values. Otherwise, the std values are a learnable nn.Parameter.
    LOG_STD_MAX : float
        Maximum std allowed value. Used for clipping.
    LOG_STD_MIN : float
        Minimum std allowed value. Used for clipping.
    """

    LOG_STD_MAX = 2
    LOG_STD_MIN = -20

    def __init__(self, num_inputs, num_outputs, predict_log_std=True):
        super(SquashedGaussian, self).__init__()

        self.predict_log_std = predict_log_std
        init_ = lambda m: init(m, nn.init.orthogonal_, lambda x: nn.init.constant_(x, 0))
        self.mean = init_(nn.Linear(num_inputs, num_outputs))

        if predict_log_std:
            self.log_std = init_(nn.Linear(num_inputs, num_outputs))
        else:
            self.log_std = nn.Parameter(torch.zeros(num_outputs).unsqueeze(0))

    def forward(self, x, deterministic=False):
        """
        Predict distribution parameters from x (obs features) and return
        actions (detemrninistic, sampled and clipped), sampled action log
        probability and distribution entropy.

        Parameters
        ----------
        x : torch.tensor
            Feature maps extracted from environment observations.
        deterministic : bool
            Whether to randomly sample action from predicted distribution or take the mode.

        Returns
        -------
        action: torch.tensor
            Predicted next action.
        clipped_action: torch.tensor
            Predicted next action (clipped to be within action space).
        logp_action : torch.tensor
            Log probability of `action` according to the predicted action distribution.
        entropy_dist : torch.tensor
            Entropy of the predicted action distribution.
        """

        # Predict distribution parameters
        action_mean = self.mean(x)
        action_logstd = self.log_std(x) if self.predict_log_std else\
            torch.zeros(action_mean.size()).to(x.device) + self.log_std
        action_logstd = torch.clamp(action_logstd, self.LOG_STD_MIN, self.LOG_STD_MAX)

        # Create distribution and sample action
        action_dist = Normal(action_mean, action_logstd.exp())

        if deterministic:
            action = action_mean
        else:
            action = action_dist.rsample()

        # Action log probability
        logp_action = action_dist.log_prob(action).sum(axis=-1, keepdim=True)
        logp_action -= (2 * (np.log(2) - action - F.softplus(-2 * action))).sum(
            axis=1, keepdim=True)

        # Apply squashing
        action = clipped_action = torch.tanh(action)

        # Distribution entropy
        entropy_dist = -logp_action.mean()

        return action, clipped_action, logp_action, entropy_dist

    def evaluate_actions(self, x, action):
        """
        Return log prob of action under the distribution generated from
        x (obs features). Also return entropy of the generated distribution.

        Parameters
        ----------
        x : torch.tensor
            obs feature map obtained from a policy_net.
        action : torch.tensor
            Evaluated action.

        Returns
        -------
        logp_action : torch.tensor
            Log probability of `action` according to the action distribution
            predicted.
        entropy_dist : torch.tensor
            Entropy of the action distribution predicted.
        """

        # Predict distribution parameters
        action_mean = self.mean(x)
        action_logstd = self.log_std(x) if self.predict_log_std else\
            torch.zeros(action_mean.size()).to(x.device) + self.log_std
        action_logstd = torch.clamp(action_logstd, self.LOG_STD_MIN, self.LOG_STD_MAX)

        # Create distribution
        action_dist = Normal(action_mean, action_logstd.exp())

        # Evaluate log prob of action under action_dist
        logp_action = action_dist.log_prob(action).sum(axis=-1, keepdim=True)
        # TODO. I think this is wrong as logp_action needs to be computed with the action previous to the tanh !!!
        # logp_action -= (2 * (np.log(2) - action - F.softplus(-2 * action))).sum(axis=1)
        logp_action -= torch.log(1 - action.pow(2) + 1e-6).sum(axis=1, keepdim=True)

        # Distribution entropy
        entropy_dist = -logp_action.mean()

        return logp_action, entropy_dist

# ---- get_dist function ------------------------------------------------------

def get_dist(name):
    """Returns model class from name."""
    if name == "Categorical":
        return Categorical
    elif name == "Gaussian":
        return DiagGaussian
    elif name == "SquashedGaussian":
        return SquashedGaussian
    else:
        raise ValueError("Specified model not found!")



import cv2
import itertools
import numpy as np


def box_is_placed(state):
    """
    Given a state returns True if the box is placed on the platform.
    Could be useful for reward scaling.
    """
    box = [132, 180, 170, 230, 78, 255]  # HSV color filters
    frame_HSV = cv2.cvtColor(state, cv2.COLOR_BGR2HSV)
    frame_threshold = cv2.inRange(
        frame_HSV, (box[0], box[2], box[4]), (box[1], box[3], box[5]))

    return (frame_threshold > 0.1).sum() > 90  # hair can appear purple


def box_location(state):
    """
    Returns a tuple: boolean 'unplaced box is visible' and 'center of it'.
    """
    box = [154, 180, 40, 102, 102, 232]  # HSV color filters
    frame_HSV = cv2.cvtColor(state, cv2.COLOR_BGR2HSV)
    frame_threshold = cv2.inRange(
        frame_HSV, (box[0], box[2], box[4]), (box[1], box[3], box[5]))
    if (frame_threshold > 0.1).sum() > 20:
        x_ord, y_ord = np.where(frame_threshold > 0.1)
        x_m = np.around(y_ord.mean()).astype(np.uint8)
        y_m = np.round(x_ord.mean()).astype(np.uint8)
        return True, [y_m, x_m]
    else:
        return False, (0, 0)


def place_location(state):
    """
    Returns a tuple: boolean 'if placing platform is visible' and 'center of it'.
    Some lighting issues might occur.
    """
    box = [0, 40, 19, 39, 92, 255]  # HSV color filters
    frame_HSV = cv2.cvtColor(state, cv2.COLOR_BGR2HSV)
    frame_threshold = cv2.inRange(
        frame_HSV, (box[0], box[2], box[4]), (box[1], box[3], box[5]))
    if (frame_threshold > 0.1).sum() > 10:
        x_ord, y_ord = np.where(frame_threshold > 0.1)
        x_m = np.around(y_ord.mean()).astype(np.uint8)
        y_m = np.round(x_ord.mean()).astype(np.uint8)
        return True, [y_m, x_m]
    else:
        return False, (0, 0)


# all actions
branched_action_space = [3, 3, 2, 3]
possible_vals = [range(_num) for _num in branched_action_space]
all_actions = [list(_action) for _action in itertools.product(*possible_vals)]
action_lookup = {tuple(_action): _scalar for (_scalar, _action) in enumerate(all_actions)}


# reduced actions
reduced_action_lookup_6 = {
    0: [0, 1, 0, 0],  # k
    1: [0, 2, 0, 0],  # l
    2: [1, 0, 0, 0],  # w
    3: [1, 0, 1, 0],  # w + space
    4: [1, 1, 0, 0],  # w + k
    5: [1, 2, 0, 0],  # w + l
}

# reduced actions
reduced_action_lookup_7 = {
    0: [0, 0, 0, 0],
    1: [0, 1, 0, 0],  # k
    2: [0, 2, 0, 0],  # l
    3: [1, 0, 0, 0],  # w
    4: [1, 0, 1, 0],  # w + space
    5: [1, 1, 0, 0],  # w + k
    6: [1, 2, 0, 0],  # w + l
}



import gym
from gym import spaces
from collections import deque
import numpy as np

class FrameSkip(gym.Wrapper):
    def __init__(self, env, skip=4):
        """Return only every `skip`-th frame"""
        gym.Wrapper.__init__(self, env)
        # most recent raw observations (for max pooling across time steps)
        self._obs_buffer = np.zeros((2,) + env.observation_space.shape, dtype=np.uint8)
        self._skip = skip

    def step(self, action):
        """Repeat action, sum reward, and max over last observations."""
        total_reward = 0.0
        done = None
        for i in range(self._skip):
            obs, reward, done, info = self.env.step(action)
            if i == self._skip - 2: self._obs_buffer[0] = obs
            if i == self._skip - 1: self._obs_buffer[1] = obs
            total_reward += reward
            if done:
                break
        # Note that the observation on the done=True frame
        # doesn't matter
        # max_frame = self._obs_buffer.max(axis=0)
        last_frame = obs

        return last_frame, total_reward, done, info

    def reset(self, **kwargs):
        return self.env.reset(**kwargs)

class FrameStack(gym.Wrapper):
    def __init__(self, env, k):
        """Stack k last frames.

        Returns lazy array, which is much more memory efficient.

        See Also
        --------
        baselines.common.atari_wrappers.LazyFrames
        """
        gym.Wrapper.__init__(self, env)
        self.k = k
        self.frames = deque([], maxlen=k)
        shp = env.observation_space.shape
        self.observation_space = spaces.Box(low=0, high=255, shape=(shp[:-1] + (shp[-1] * k,)), dtype=env.observation_space.dtype)

    def reset(self):
        ob = self.env.reset()
        for _ in range(self.k):
            self.frames.append(ob)
        return self._get_ob()

    def step(self, action):
        ob, reward, done, info = self.env.step(action)
        self.frames.append(ob)
        return self._get_ob(), reward, done, info

    def _get_ob(self):
        assert len(self.frames) == self.k
        # return LazyFrames(list(self.frames))
        return np.concatenate(self.frames, axis=-1)


import ray
import time
import torch
from ..base.worker import Worker as W
from ..utils import TaskPool, ray_get_and_free


class GWorker(W):
    """
    Worker class handling gradient computation.

    This class wraps an actor_critic instance, a storage class instance and a
    worker set of remote data collection workers. It receives data from the
    collection workers and computes gradients following a logic defined in
    function self.step(), which will be called from the Learner class.

    Parameters
    ----------
    index_worker : int
        Worker index.
    create_algo_instance : func
        A function that creates an algorithm class.
    create_storage_instance : func
        A function that create a rollouts storage.
    create_actor_critic_instance : func
        A function that creates a policy.
    initial_weights : ray object ID
        Initial model weights.
    max_collect_requests_pending : int
        maximum number of collection tasks simultaneously scheduled to each
        collection worker.

    Attributes
    ----------
    index_worker : int
        Index assigned to this worker.
    actor_critic : nn.Module
        An actor_critic class instance.
    algo : an algorithm class
        An algorithm class instance.
    storage : a rollout storage class
        A Storage class instance.
    iter : int
        Times gradients have been computed and sent.
    ac_version : int
        Number of times the current actor critic version been has been updated.
    latest_weights : ray object ID
        Last received model weights.
    collector_tasks : TaskPool
        Tracks the status of in-flight actor collection tasks.
    """

    def __init__(self,
                 index_worker,
                 create_algo_instance,
                 create_storage_instance,
                 create_actor_critic_instance,
                 create_collection_worker_set_instance,
                 max_collect_requests_pending=2,
                 initial_weights=None):

        super(GWorker, self).__init__(index_worker)

        # worker should only see one GPU or None
        device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

        # Create Actor Critic instance
        self.actor_critic = create_actor_critic_instance(device)
        self.actor_critic.to(device)

        # Create Algorithm instance
        self.algo = create_algo_instance(self.actor_critic, device)

        # Define counters and other attributes
        self.iter = 0
        self.ac_version = 0

        if initial_weights:  # if remote worker

            # Create CWorkerSet instance
            self.c_workers = create_collection_worker_set_instance(
                initial_weights, create_algo_instance, create_storage_instance,
                ).remote_workers()

            # Create Storage instance and set world initial state
            self.storage = create_storage_instance(device)

            # Print worker information
            self.print_worker_info()

            # Set initial weights to collector workers and start data collection
            self.latest_weights = initial_weights
            self.collector_tasks = TaskPool()
            for collector_worker in self.c_workers:
                for _ in range(max_collect_requests_pending):
                    collector_worker.set_weights.remote(self.latest_weights)
                    self.collector_tasks.add(
                        collector_worker, collector_worker.collect_data.remote())

    def compute_gradients(self, batch):
        """
        Calculate actor critic gradients and update networks.

        Parameters
        ----------
        batch : dict
            data batch containing all required tensors to compute algo loss.

        Returns
        -------
        grads: list of tensors
            List of actor_critic gradients.
        info : dict
            Summary dict with relevant gradient-related information.
        """

        t = time.time()
        grads, info = self.algo.compute_gradients(batch)
        info.update({"scheme/seconds_to/compute_grads": time.time() - t})

        return grads, info

    def update_networks(self, gradients):
        """Update Actor Critic model"""
        self.algo.apply_gradients(gradients)

    def step(self):
        """
        Perform logical learning step. Training proceeds receiving data samples
        from collection workers and computations policy gradients.

        Returns
        -------
        grads: list of tensors
            List of actor_critic gradients.
        info : dict
            Summary dict of relevant step information.
        """

        # Collect data and prepare data batches
        if self.iter % (self.algo.num_epochs * self.algo.num_mini_batch) == 0:

            # Wait to remote workers to complete data collection tasks
            for collector_worker, rollouts in self.collector_tasks.completed(blocking_wait=True, max_yield=1):
                new_rollouts = ray_get_and_free(rollouts)
                collector_worker.set_weights.remote(self.latest_weights)
                self.collector_tasks.add(collector_worker, collector_worker.collect_data.remote())

            self.collect_info = new_rollouts["info"]
            self.storage.add_data(new_rollouts["data"])
            self.storage.before_update(self.actor_critic, self.algo)
            self.batches = self.storage.generate_batches(
                self.algo.num_mini_batch, self.algo.mini_batch_size,
                self.algo.num_epochs, self.actor_critic.is_recurrent)

        else:
            collect_info = {}
            collect_info.update({"collected_samples": 0})

        # Compute gradients, get algo info
        grads, info = self.compute_gradients(self.batches.__next__())

        # Add extra information to info dict
        info.update(self.collect_info)
        self.collect_info.update({"collected_samples": 0})
        info.update({"ac_update_num": self.iter})
        info.update({"scheme/metrics/collection_gradient_delay": self.iter - self.collect_info["ac_version"]})

        # Update counter
        self.iter += 1

        return grads, info

    def set_weights(self, weights):
        """
        Update the worker actor_critic version with provided weights.

        weights: dict of tensors
            Dict containing actor_critic weights to be set.
        """
        self.latest_weights = weights
        self.ac_version = weights["update"]
        self.algo.set_weights(weights["weights"])

    def terminate_worker(self):
        """Terminate data collection remote workers"""
        for e in self.c_workers:
            e.terminate_worker.remote()
        ray.actor.exit_actor()

    def update_algo_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of Worker.algo, change its value to
        `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Algorithm attribute name
        """
        self.algo.update_algo_parameter(parameter_name, new_parameter_value)
        for e in self.c_workers.remote_workers():
            e.update_algo_parameter.remote(parameter_name, new_parameter_value)


""" Neural Network Architectures. """

import math
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from .utils import init

# ---- Neural Network Base Class ----------------------------------------------

class NNBase(nn.Module):
    """
    Base Neural Network class.

    Parameters
    ----------
    input_shapes : list of tuples
        List of all input shapes.
    output_shape : tuple
        Shape of output feature map.
    feature_extractor : nn.Module
        PyTorch nn.Module used as the features extraction block.
    feature_extractor_kwargs : dict
        Keyword arguments for the feature extractor network.
    recurrent : bool
        Whether to use recurrency or not.
    recurrent_input_size (int):
        Input size of GRU layer.
    recurrent_hidden_size (int):
        Hidden size of GRU layer.
    activation : func
        Non-linear activation function.
    final_activation : bool
        Whether or not to apply nonlinearity after last layer.

    Attributes
    ----------
    is_recurrent : bool
        True if this is a recurrent neural network.
    num_outputs : int
        Output feature map size (as in np.prod(self.output_shape)).
    output_shape : tuple
        Output feature map shape.
    recurrent_hidden_state_size : int
        Recurrent hidden state size
    feature_extractor : nn.Module
        Neural network feature extractor block.
    gru : nn.Module
        Neural network recurrency block.
    output : nn.Module
        Neural network final layer.
    """
    def __init__(self,
                 input_shapes,
                 feature_extractor,
                 feature_extractor_kwargs,
                 recurrent=False,
                 output_shape=(256,),
                 recurrent_hidden_size=512,
                 activation=nn.ReLU,
                 final_activation=False):

        super(NNBase, self).__init__()

        self.output_shape = output_shape
        self._num_outputs = np.prod(output_shape)
        self._recurrent = recurrent
        self._recurrent_hidden_size = recurrent_hidden_size

        if len(input_shapes) == 1: obs_shape, action_shape = input_shapes[0], 0
        elif len(input_shapes) == 2: obs_shape, action_shape = input_shapes
        else: raise NotImplementedError

        self.feature_extractor = feature_extractor(obs_shape, **feature_extractor_kwargs)
        feature_map_size = int(np.prod(self.feature_extractor(torch.randn(1, *obs_shape)).shape))

        if recurrent:
            self.gru = nn.GRU(feature_map_size, recurrent_hidden_size)
            feature_map_size = recurrent_hidden_size

        output = [nn.Linear(feature_map_size + np.prod(action_shape), self._num_outputs)]
        if final_activation: output += [activation()]
        self.output = nn.Sequential(*output)

        self.train()

    @property
    def is_recurrent(self):
        """Returns True if this is a recurrent neural network"""
        return self._recurrent

    @property
    def num_outputs(self):
        """Returns output feature map size"""
        return self._num_outputs

    @property
    def recurrent_hidden_state_size(self):
        """Returns recurrent hidden state size"""
        return self._recurrent_hidden_size

    def initial_states(self, obs):
        """
        Returns all policy inputs to predict the environment initial action.

        Parameters
        ----------
        obs : torch.tensor
            Initial environment observation.

        Returns
        -------
        obs : torch.tensor
            Initial environment observation.
        rhs : torch.tensor
            Initial recurrent hidden state (will contain zeroes).
        done : torch.tensor
            Initial done tensor, indicating the environment is not done.
        """
        done = torch.zeros(obs.size(0), 1).to(obs.device)
        rhs = torch.zeros(obs.size(0), self._recurrent_hidden_size).to(obs.device)
        return obs, rhs, done

    def _forward_gru(self, x, hxs, done):
        """
        Fast forward pass GRU network.
        from Ilya Kostrikov.PyTorch Implementations of Reinforcement Learning Algorithms.
        https://github.com/ikostrikov/pytorch-a2c-ppo-acktr-gail. 2018

        Parameters
        ----------
        x : torch.tensor
            Feature map obtained from environment observation.
        hxs : torch.tensor
            Current recurrent hidden state.
        done : torch.tensor
            Current done tensor, indicating if episode has finished.

        Returns
        -------
        x : torch.tensor
            Feature map obtained after GRU.
        hxs : torch.tensor
            Updated recurrent hidden state.
        """

        masks = 1 - done

        if x.size(0) == hxs.size(0):
            x, hxs = self.gru(x.unsqueeze(0), (hxs * masks).unsqueeze(0))
            x = x.squeeze(0)
            hxs = hxs.squeeze(0)
        else:
            # x is a (T, N, -1) tensor that has been flatten to (T * N, -1)
            N = hxs.size(0)
            T = int(x.size(0) / N)

            # unflatten
            x = x.view(T, N, x.size(1))

            # Same deal with masks
            masks = masks.view(T, N)

            # Let's figure out which steps in the sequence have a zero for any agent
            # We will always assume t=0 has a zero in it as that makes the logic cleaner
            has_zeros = ((masks[1:] == 0.0) \
                         .any(dim=-1)
                         .nonzero()
                         .squeeze()
                         .cpu())

            # +1 to correct the masks[1:]
            if has_zeros.dim() == 0:
                # Deal with scalar
                has_zeros = [has_zeros.item() + 1]
            else:
                has_zeros = (has_zeros + 1).numpy().tolist()

            # add t=0 and t=T to the list
            has_zeros = [0] + has_zeros + [T]

            hxs = hxs.unsqueeze(0)
            outputs = []
            for i in range(len(has_zeros) - 1):
                # We can now process steps that don't have any zeros in masks together!
                # This is much faster
                start_idx = has_zeros[i]
                end_idx = has_zeros[i + 1]

                rnn_scores, hxs = self.gru(
                    x[start_idx:end_idx],
                    hxs * masks[start_idx].view(1, -1, 1))

                outputs.append(rnn_scores)

            # x is a (T, N, -1) tensor
            x = torch.cat(outputs, dim=0)
            # flatten
            x = x.view(T * N, -1)
            hxs = hxs.squeeze(0)

        return x, hxs

    def forward(self, inputs, hxs=None, done=None):
        """
        Forward pass Neural Network

        Parameters
        ----------
        inputs : torch.tensor or tuple
            Input data, containing obs or obs + action.
        hxs : torch.tensor
            Current recurrent hidden state.
        done : torch.tensor
            Current done tensor, indicating if episode has finished.

        Returns
        -------
        out : torch.tensor
            Output feature map.
        hxs : torch.tensor
            Updated recurrent hidden state.
        """

        if isinstance(inputs, tuple):
            obs, act = inputs
        else:
            obs, act = inputs, None

        x = self.feature_extractor(obs)
        x = x.view(x.size(0), -1)

        if self.is_recurrent:
            x, rnn_hxs = self._forward_gru(x, hxs, done)

        if act is not None:
            x = torch.cat([x, act], dim=-1)

        out = self.output(x)
        assert tuple(out.shape[1:]) == self.output_shape

        return out, hxs

# ---- Multilayer Perceptron --------------------------------------------------

class MLP(nn.Module):
    """
    Multilayer Perceptron network

    Parameters
    ----------
    input_shape : tuple
        Shape input tensors.
    hidden_sizes : list
        Hidden layers sizes.
    activation : func
        Non-linear activation function.

    Attributes
    ----------
    feature_extractor : nn.Module
        Neural network feature extractor block.
    """
    def __init__(self, input_shape, hidden_sizes=[256, 256], activation=nn.ReLU):
        super(MLP, self).__init__()

        init_ = lambda m: init(m, nn.init.orthogonal_, lambda x: nn.init. constant_(x, 0), nn.init.calculate_gain('relu'))

        # Define feature extractor
        layers = []
        sizes = [input_shape[0]] + hidden_sizes
        for j in range(len(sizes) - 1):
            layers += [init_(nn.Linear(sizes[j], sizes[j + 1])), activation()]
        self.feature_extractor = nn.Sequential(*layers)

        self.train()


    def forward(self, inputs):
        """
        Forward pass Neural Network

        Parameters
        ----------
        inputs : torch.tensor
            Input data.

        Returns
        -------
        out : torch.tensor
            Output feature map.
        """
        out = self.feature_extractor(inputs)
        return out

# ---- Basic Convolutional Neural Network -------------------------------------

class CNN(nn.Module):
    """
    Convolutional Neural Network.

    Parameters
    ----------
    input_shape : tuple
        Shape input tensors.
    activation : func
        Non-linear activation function.
    strides : list
        Convolutional layers strides.
    filters : list
        Convolutional layers number of filters.
    kernel_sizes : list
        Convolutional layers kernel sizes.

    Attributes
    ----------
    feature_extractor : nn.Module
        Neural network feature extractor block.
    """
    def __init__(self,
                 input_shape,
                 activation=nn.ReLU,
                 strides=[4, 2, 1],
                 filters=[32, 64, 64],
                 kernel_sizes=[8, 4, 3]):
        super(CNN, self).__init__()

        assert len(filters) == len(strides) and len(strides) == len(kernel_sizes)

        init_ = lambda m: init(m, nn.init.orthogonal_, lambda x: nn.init.constant_(x, 0), nn.init.calculate_gain('relu'))

        # Define feature extractor
        layers = []
        filters = [input_shape[0]] + filters
        for j in range(len(filters) - 1):
            layers += [init_(nn.Conv2d(
                filters[j], filters[j + 1], stride=strides[j],
                kernel_size=kernel_sizes[j])), activation()]
        self.feature_extractor = nn.Sequential(*layers)
        self.train()

    def forward(self, inputs):
        """
        Forward pass Neural Network

        Parameters
        ----------
        inputs : torch.tensor
            Input data.

        Returns
        -------
        out : torch.tensor
            Output feature map.
        """
        out = self.feature_extractor(inputs / 255.0)
        return out

# ---- Fixup Convolutional Neural Network -------------------------------------

class FixupCNN(nn.Module):
    """
    A larger version of the IMPALA CNN with Fixup init.
    See Fixup: https://arxiv.org/abs/1901.09321.

    Parameters
    ----------
    input_shape : tuple
        Shape input tensors.

    Attributes
    ----------
    feature_extractor : nn.Module
        Neural network feature extractor block.
    """
    def __init__(self, input_shape):
        super(FixupCNN, self).__init__()

        depth_in = input_shape[0]
        layers = []
        for depth_out in [16, 32, 32]:
            layers.extend([
                nn.Conv2d(depth_in, depth_out, 3, padding=1),
                nn.MaxPool2d(3, stride=2, padding=1),
                FixupResidualModule(depth_out, 6),
                FixupResidualModule(depth_out, 6),
            ])
            depth_in = depth_out
        self.feature_extractor = nn.Sequential(*layers)
        self.train()

    def forward(self, inputs):
        """
        Forward pass Neural Network

        Parameters
        ----------
        inputs : torch.tensor
            Input data.

        Returns
        -------
        out : torch.tensor
            Output feature map.
        """
        out = self.feature_extractor(inputs / 255.0)
        return out

class FixupResidualModule(nn.Module):
    """
    FixupCNN resudial block.

    Parameters
    ----------
    depth : int
        Number of input feature maps.
    num_residual : int
        Number of residual feature blocks in the architecture
    """
    def __init__(self, depth, num_residual):
        super(FixupResidualModule, self).__init__()

        self.conv1 = nn.Conv2d(depth, depth, 3, padding=1, bias=False)
        self.conv2 = nn.Conv2d(depth, depth, 3, padding=1, bias=False)

        for p in self.conv1.parameters():
            p.data.mul_(1 / math.sqrt(num_residual))
        for p in self.conv2.parameters():
            p.data.zero_()

        self.bias1 = nn.Parameter(torch.zeros([depth, 1, 1]))
        self.bias2 = nn.Parameter(torch.zeros([depth, 1, 1]))
        self.bias3 = nn.Parameter(torch.zeros([depth, 1, 1]))
        self.bias4 = nn.Parameter(torch.zeros([depth, 1, 1]))
        self.scale = nn.Parameter(torch.ones([depth, 1, 1]))

    def forward(self, x):
        """
        Forward pass residual block

        Parameters
        ----------
        inputs : torch.tensor
            Input feature map.

        Returns
        -------
        out : torch.tensor
            Output feature map.
        """
        x = F.relu(x)
        out = x + self.bias1
        out = self.conv1(out)
        out = out + self.bias2
        out = F.relu(out)
        out = out + self.bias3
        out = self.conv2(out)
        out = out * self.scale
        out = out + self.bias4
        return out + x

# ---- get_model function -----------------------------------------------------

def get_model(name):
    """Returns model class from name."""
    if name == "MLP":
        return MLP
    elif name == "CNN":
        return CNN
    elif name == "Fixup":
        return FixupCNN
    else:
        raise ValueError("Specified model not found!")




import itertools
import numpy as np
from copy import deepcopy

import torch
import torch.optim as optim

from .base import Algo


class SAC(Algo):
    """
    Soft Actor Critic algorithm class.

    Algorithm class to execute SAC, from Haarnoja et al.
    (https://arxiv.org/abs/1812.05905). Algorithms are modules generally
    required by multiple workers, so SAC.algo_factory(...) returns a function
    that can be passed on to workers to instantiate their own SAC module.

    Parameters
    ----------
    device : torch.device
        CPU or specific GPU where class computations will take place.
    actor_critic : ActorCritic
        Actor_critic class instance.
    lr : float
        Optimizers learning rate.
    eps : float
        Optimizers epsilon parameter.
    gamma : float
        Discount factor parameter.
    alpha: float
        Initial entropy coefficient value.
    polyak: float
        SAC polyak averaging parameter.
    num_updates: int
        Num consecutive actor_critic updates before data collection continues.
    update_every: int
        Regularity of actor_critic updates in number environment steps.
    start_steps: int
        Num of initial random environment steps before learning starts.
    mini_batch_size: int
        Size of actor_critic update batches.
    reward_scaling: float
        Reward scaling factor.
    target_update_interval: float
        regularity of target nets updates with respect to actor_critic Adam updates.
    num_test_episodes : int
        Number of episodes to complete in each test phase.
    test_every : int
        Regularity of test evaluations in actor_critic updates.

    Attributes
    ----------
    start_steps : int
        Number of steps collected with initial random policy (default to 0 for
        On-policy algos).
    num_epochs : int
        Times data in the buffer is re-used before data collection proceeds.
    update_every : int
        Number of data samples collected between network update stages (depends
        on storage capacity for On-policy algos).
    num_mini_batch : int
        Number mini batches per epoch.
    mini_batch_size : int
        Size of update mini batches.
    test_every : int
        Number of network updates between test evaluations.
    num_test_episodes : int
        Num episodes to complete in each test phase.
    device : torch.device
        CPU or specific GPU where class computation take place.
    actor_critic : ActorCritic
        ActorCritic Class containing Neural Network function approximators.
    gamma : float
        Discount factor parameter.
    iter : int
        Num actor_critic Adam updates.
    polyak : float
        SAC polyak averaging parameter.
    reward_scaling : float
        Reward scaling factor.
    target_update_interval : int
        regularity of target nets updates with respect to actor_critic Adam updates.
    log_alpha : torch.tensor
        Log entropy coefficient value.
    alpha : torch.tensor
        Entropy coefficient value.
    pi_optimizer : torch.optimizer
        Policy model optimizer.
    q_optimizer : torch.optimizer
        Q critics model optimizer.
    alpha_optimizer : torch.optimizer
        alpha parameter optimizer.

    Examples
    --------
    >>> ac = ActorCritic(...)
    >>> create_algo = SAC.algo_factory(
            lr=1e-4, eps=1e-5, gamma=0.99, polyak=0.995, num_updates=50,
            update_every=50, test_every=5000, start_steps=20000,
            mini_batch_size=64, alpha=1.0, reward_scaling=1.0,
            num_test_episodes=0, target_update_interval=1)
    >>> sac1 = create_algo(actor_critic=ac, device=torch.device("cuda:1"))
    >>> sac2 = create_algo(actor_critic=ac, device=torch.device("cuda:2"))

    >>> ac = ActorCritic(...)
    >>> sac = SAC(device=torch.device("cuda:1"), actor_critic=ac, lr=1e-4,
                  eps=1e-5, gamma=0.99, polyak=0.995, num_updates=50,
                  update_every=50, test_every=5000, start_steps=20000,
                  mini_batch_size=64, alpha=1.0, reward_scaling=1.0,
                  num_test_episodes=0, target_update_interval=1)
    """

    def __init__(self,
                 device,
                 actor_critic,
                 lr=1e-4,
                 eps=1e-5,
                 alpha=1.0,
                 gamma=0.99,
                 polyak=0.995,
                 num_updates=50,
                 update_every=50,
                 test_every=5000,
                 start_steps=20000,
                 mini_batch_size=64,
                 reward_scaling=1.0,
                 num_test_episodes=0,
                 target_update_interval=1):

        # ---- General algo attributes ----------------------------------------

        # Number of steps collected with initial random policy
        self.start_steps = start_steps

        # Times data in the buffer is re-used before data collection proceeds
        self.num_epochs = 1 # Default to 1 for off-policy algorithms

        # Number of data samples collected between network update stages
        self.update_every = update_every

        # Number mini batches per epoch
        self.num_mini_batch = num_updates

        # Size of update mini batches
        self.mini_batch_size = mini_batch_size

        # Number of network updates between test evaluations
        self.test_every = test_every

        # Number of episodes to complete when testing
        self.num_test_episodes = num_test_episodes

        # ---- SAC-specific attributes ----------------------------------------

        self.iter = 0
        self.gamma = gamma
        self.polyak = polyak
        self.device = device
        self.actor_critic = actor_critic
        self.reward_scaling = reward_scaling
        self.target_update_interval = target_update_interval

        if self.actor_critic.q1 is None:
            raise ValueError("SAC requires double q critic")

        self.log_alpha = torch.tensor(
            data=[np.log(alpha)], dtype=torch.float32,
            requires_grad=True, device=device)
        self.alpha = self.log_alpha.detach().exp()

        # Compute target entropy
        target_entropy = self.calculate_target_entropy()
        self.target_entropy = torch.tensor(
            data=target_entropy, dtype=torch.float32,
            requires_grad=False, device=device)

        # Create target networks
        self.actor_critic_targ = deepcopy(actor_critic)

        # Freeze target networks with respect to optimizers (only update via polyak averaging)
        for p in self.actor_critic_targ.parameters():
            p.requires_grad = False

        # List of parameters for both Q-networks
        q_params = itertools.chain(self.actor_critic.q1.parameters(),
                                   self.actor_critic.q2.parameters())

        # List of parameters for both Q-networks
        p_params = itertools.chain(self.actor_critic.policy_net.parameters(),
                                   self.actor_critic.dist.parameters())

        # ----- Optimizers ----------------------------------------------------

        self.pi_optimizer = optim.Adam(p_params, lr=lr, eps=eps)
        self.q_optimizer = optim.Adam(q_params, lr=lr, eps=eps)
        self.alpha_optimizer = optim.Adam([self.log_alpha], lr=lr, eps=eps)

    @classmethod
    def algo_factory(cls,
                     lr=1e-4,
                     eps=1e-5,
                     alpha=1.0,
                     gamma=0.99,
                     polyak=0.995,
                     num_updates=50,
                     update_every=50,
                     start_steps=20000,
                     mini_batch_size=64,
                     reward_scaling=1.0,
                     target_update_interval=1.0):
        """
        Returns a function to create new SAC instances.

        Parameters
        ----------
        lr : float
            Optimizers learning rate.
        eps : float
            Optimizers epsilon parameter.
        gamma : float
            Discount factor parameter.
        alpha: float
            Initial entropy coefficient value.
        polyak: float
            SAC polyak averaging parameter.
        num_updates: int
            Num consecutive actor_critic updates before data collection continues.
        update_every: int
            Regularity of actor_critic updates in number environment steps.
        start_steps: int
            Num of initial random environment steps before learning starts.
        mini_batch_size: int
            Size of actor_critic update batches.
        reward_scaling: float
            Reward scaling factor.
        target_update_interval: float
            regularity of target nets updates with respect to actor_critic Adam updates.
        num_test_episodes : int
            Number of episodes to complete in each test phase.
        test_every : int
            Regularity of test evaluations in actor_critic updates.

        Returns
        -------
        create_algo_instance : func
            creates a new SAC class instance.
        """

        def create_algo_instance(actor_critic, device):
            return cls(lr=lr,
                       eps=eps,
                       gamma=gamma,
                       alpha=alpha,
                       device=device,
                       polyak=polyak,
                       start_steps=start_steps,
                       num_updates=num_updates,
                       update_every=update_every,
                       actor_critic=actor_critic,
                       reward_scaling=reward_scaling,
                       mini_batch_size=mini_batch_size,
                       target_update_interval=target_update_interval)
        return create_algo_instance

    @property
    def discrete_version(self):
        """Returns True if action_space is discrete."""
        return self.actor_critic.action_space.__class__.__name__ == "Discrete"

    def acting_step(self, obs, rhs, done, deterministic=False):
        """
        SAC acting function.

        Parameters
        ----------
        obs: torch.tensor
            Current world observation
        rhs: torch.tensor
            RNN recurrent hidden state (if policy is not a RNN, rhs will contain zeroes).
        done: torch.tensor
            1.0 if current obs is the last one in the episode, else 0.0.
        deterministic: bool
            Whether to randomly sample action from predicted distribution or taking the mode.

        Returns
        -------
        action: torch.tensor
            Predicted next action.
        clipped_action: torch.tensor
            Predicted next action (clipped to be within action space).
        rhs: torch.tensor
            Policy recurrent hidden state (if policy is not a RNN, rhs will contain zeroes).
        other: dict
            Additional SAC predictions, which are not used in other algorithms.
        """

        with torch.no_grad():
            (action, clipped_action, logp_action, rhs,
             entropy_dist) = self.actor_critic.get_action(
                obs, rhs, done, deterministic=deterministic)

        return action, clipped_action, rhs, {}

    def compute_loss_q(self, data, rnn_hs):
        """
        Calculate SAC Q-nets loss

        Parameters
        ----------
        data: dict
            Data batch dict containing all required tensors to compute SAC losses.
        rnn_hs : torch.tensor
            Policy recurrent hidden state obtained with the current ActorCritic version.

        Returns
        -------
        loss_q1 : torch.tensor
            Q1-net loss.
        loss_q2 : torch.tensor
            Q2-net loss.
        loss_q : torch.tensor
            Weighted average of loss_q1 and loss_q2.
        """

        o, a, r, o2, d = data["obs"], data["act"], data["rew"], data["obs2"], data["done"]
        r *= self.reward_scaling

        if self.discrete_version:

            # Q-values for all actions
            q1, q2 = self.actor_critic.get_q_scores(o)
            q1 = q1.gather(1, a.long())
            q2 = q2.gather(1, a.long())

            # Bellman backup for Q functions
            with torch.no_grad():
                # Target actions come from *current* policy
                a2, _, _, _, _ = self.actor_critic.get_action(o2, rnn_hs, d)
                p_a2 = self.actor_critic.dist.action_dist.probs
                z = (p_a2 == 0.0).float() * 1e-8
                logp_a2 = torch.log(p_a2 + z)

                # Target Q-values
                q1_pi_targ, q2_pi_targ = self.actor_critic_targ.get_q_scores(o2)
                q_pi_targ = (p_a2 * (torch.min(q1_pi_targ, q2_pi_targ) - self.alpha * logp_a2)).sum(dim=1, keepdim=True)

                assert r.shape == q_pi_targ.shape
                backup = r + self.gamma * (1 - d) * q_pi_targ

        else:

            # Q-values for all actions
            q1, q2 = self.actor_critic.get_q_scores(o, a)

            # Bellman backup for Q functions
            with torch.no_grad():

                # Target actions come from *current* policy
                a2, _, logp_a2, _, _ = self.actor_critic.get_action(o2, rnn_hs, d)

                # Target Q-values
                q1_pi_targ, q2_pi_targ = self.actor_critic_targ.get_q_scores(o2, a2)
                q_pi_targ = torch.min(q1_pi_targ, q2_pi_targ)

                backup = r + self.gamma * (1 - d) * (q_pi_targ - self.alpha * logp_a2)

        # MSE loss against Bellman backup
        loss_q1 = ((q1 - backup) ** 2).mean()
        loss_q2 = ((q2 - backup) ** 2).mean()
        loss_q = 0.5 * loss_q1 + 0.5 * loss_q2

        return loss_q1, loss_q2, loss_q

    def compute_loss_pi(self, data):
        """
        Calculate SAC policy loss.

        Parameters
        ----------
        data: dict
            Data batch dict containing all required tensors to compute SAC losses.

        Returns
        -------
        loss_pi : torch.tensor
            SAC policy loss.
        logp_pi : torch.tensor
            Log probability of predicted next action.
        rnn_hs : torch.tensor
            Policy recurrent hidden state obtained with the current ActorCritic version.
        """

        o, rhs, a, r, o2, d = data["obs"], data["rhs"], data["act"], data["rew"], data["obs2"], data["done"]
        r *= self.reward_scaling

        if self.discrete_version:

            pi, _, _, rnn_hs, _ = self.actor_critic.get_action(o, rhs, d)
            p_pi = self.actor_critic.dist.action_dist.probs
            z = (p_pi == 0.0).float() * 1e-8
            logp_pi = torch.log(p_pi + z)
            logp_pi = torch.sum(p_pi * logp_pi, dim=1, keepdim=True)
            q1_pi, q2_pi = self.actor_critic.get_q_scores(o)
            q_pi = torch.sum(torch.min(q1_pi, q2_pi) * p_pi, dim=1, keepdim=True)

        else:

            pi, _, logp_pi, rnn_hs, _ = self.actor_critic.get_action(o, rhs, d)
            q1_pi, q2_pi = self.actor_critic.get_q_scores(o, pi)
            q_pi = torch.min(q1_pi, q2_pi)

        loss_pi = (self.alpha * logp_pi - q_pi).mean()
        return loss_pi, logp_pi, rnn_hs

    def compute_loss_alpha(self, log_probs):
        """
        Calculate SAC entropy loss.

        Parameters
        ----------
        log_probs: torch.tensor

        Returns
        -------
        alpha_loss: torch.tensor
            SAC entropy loss.
        """
        alpha_loss = - (self.log_alpha * (log_probs + self.target_entropy).detach()).mean()
        return alpha_loss

    def calculate_target_entropy(self):
        """Calculate SAC target entropy"""
        if self.discrete_version:
            target = - np.log(1.0 / self.actor_critic.action_space.n) * 0.98
        else:
            target = - self.actor_critic.action_space.shape[0]
        return target

    def compute_gradients(self, batch, grads_to_cpu=True):
        """
        Compute loss and compute gradients but don't do optimization step,
        return gradients instead.

        Parameters
        ----------
        data: dict
            data batch containing all required tensors to compute SAC losses.
        grads_to_cpu: bool
            If gradient tensor will be sent to another node, need to be in CPU.

        Returns
        -------
        grads: list of tensors
            List of actor_critic gradients.
        info: dict
            Dict containing current SAC iteration information.
        """

        # Compute policy and Q losses
        loss_pi, logp_pi, rhs = self.compute_loss_pi(batch)
        loss_q1, loss_q2, loss_q = self.compute_loss_q(batch, rhs)

        # First run one gradient descent step for Q1 and Q2
        self.q_optimizer.zero_grad()
        loss_q.backward()

        for p in itertools.chain(self.actor_critic.q1.parameters(), self.actor_critic.q2.parameters()):
            p.requires_grad = False

        # Next run one gradient descent step for pi.
        self.pi_optimizer.zero_grad()
        loss_pi.backward()

        for p in itertools.chain(self.actor_critic.q1.parameters(), self.actor_critic.q2.parameters()):
            p.requires_grad = True

        grads = []
        for p in self.actor_critic.parameters():
            if grads_to_cpu:
                if p.grad is not None:
                    grads.append(p.grad.data.cpu().numpy())
                else:
                    grads.append(None)
            else:
                if p.grad is not None:
                    grads.append(p.grad)

        # Finally, next run one gradient descent step for alpha.
        self.alpha_optimizer.zero_grad()
        loss_alpha = self.compute_loss_alpha(logp_pi)
        loss_alpha.backward()

        info = {
            "algo/loss_q1": loss_q1.detach().item(),
            "algo/loss_q2": loss_q2.detach().item(),
            "algo/loss_pi": loss_pi.detach().item(),
            "algo/loss_alpha": loss_alpha.detach().item(),
            "algo/alpha": self.alpha.detach().item(),
        }

        return grads, info

    def update_target_networks(self):
        """Update actor critic target networks with polyak averaging"""
        if self.iter % self.target_update_interval == 0:
            with torch.no_grad():
                for p, p_targ in zip(self.actor_critic.parameters(), self.actor_critic_targ.parameters()):
                    p_targ.data.mul_(self.polyak)
                    p_targ.data.add_((1 - self.polyak) * p.data)

    def apply_gradients(self, gradients=None):
        """
        Take an optimization step, previously setting new gradients if provided.
        Update also target networks.

        Parameters
        ----------
        gradients: list of tensors
            List of actor_critic gradients.
        """
        if gradients is not None:
            for g, p in zip(gradients, self.actor_critic.parameters()):
                if g is not None:
                    p.grad = torch.from_numpy(g).to(self.device)

        self.q_optimizer.step()
        self.pi_optimizer.step()
        self.alpha_optimizer.step()
        self.alpha = self.log_alpha.detach().exp()

        # Update target networks by polyak averaging.
        self.iter += 1
        self.update_target_networks()

    def set_weights(self, weights):
        """
        Update actor critic with the given weights. Update also target networks.

        Parameters
        ----------
        weights: dict of tensors
            Dict containing actor_critic weights to be set.
        """
        self.actor_critic.load_state_dict(weights)
        self.alpha_optimizer.step()
        self.alpha = self.log_alpha.detach().exp()

        # Update target networks by polyak averaging.
        self.iter += 1
        self.update_target_networks()

    def update_algo_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of the algorithm, change its value
        to `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Worker.algo attribute name
        new_parameter_value : int or float
            New value for `parameter_name`.
        """
        if hasattr(self, parameter_name):
            setattr(self, parameter_name, new_parameter_value)
        if parameter_name == "lr":
            for param_group in self.pi_optimizer.param_groups:
                param_group['lr'] = new_parameter_value
            for param_group in self.q_optimizer.param_groups:
                param_group['lr'] = new_parameter_value
            for param_group in self.alpha_optimizer.param_groups:
                param_group['lr'] = new_parameter_value


import os
os.environ.setdefault('PATH', '')
import gym
import cv2
import numpy as np
cv2.ocl.setUseOpenCL(False)
from ..common import FrameStack


class FireResetEnv(gym.Wrapper):
    def __init__(self, env):
        """Take action on reset for environments that are fixed until firing."""
        gym.Wrapper.__init__(self, env)
        assert env.unwrapped.get_action_meanings()[1] == 'FIRE'
        assert len(env.unwrapped.get_action_meanings()) >= 3

    def reset(self, **kwargs):
        self.env.reset(**kwargs)
        obs, _, done, _ = self.env.step(1)
        if done:
            self.env.reset(**kwargs)
        obs, _, done, _ = self.env.step(2)
        if done:
            self.env.reset(**kwargs)
        return obs

    def step(self, ac):
        return self.env.step(ac)

class EpisodicLifeEnv(gym.Wrapper):
    def __init__(self, env):
        """Make end-of-life == end-of-episode, but only reset on true game over.
        Done by DeepMind for the DQN and co. since it helps value estimation.
        """
        gym.Wrapper.__init__(self, env)
        self.lives = 0
        self.was_real_done  = True
        self.total_reward = 0.0

    def step(self, action):
        obs, reward, done, info = self.env.step(action)

        self.total_reward += reward
        info['rrr'] = self.total_reward

        self.was_real_done = done
        # check current lives, make loss of life terminal,
        # then update lives to handle bonus lives
        lives = self.env.unwrapped.ale.lives()
        if lives < self.lives and lives > 0:
            # for Qbert sometimes we stay in lives == 0 condition for a few frames
            # so it's important to keep lives > 0, so that we only reset once
            # the environment advertises done.
            done = True
        info['lives'] = lives
        self.lives = lives
        return obs, reward, done, info

    def reset(self, **kwargs):
        """Reset only when lives are exhausted.
        This way all states are still reachable even though lives are episodic,
        and the learner need not know about any of this behind-the-scenes.
        """
        if self.was_real_done:
            self.total_reward = 0.0
            obs = self.env.reset(**kwargs)
        else:
            # no-op step to advance from terminal/lost life state
            obs, _, _, _ = self.env.step(0)
        self.lives = self.env.unwrapped.ale.lives()
        return obs

class ClipRewardEnv(gym.Wrapper):
    def __init__(self, env):
        gym.Wrapper.__init__(self, env)
        self.total_reward = 0.0

    def step(self, action):
        obs, reward, done, info = self.env.step(action)

        self.total_reward += reward
        info['rr'] = self.total_reward

        reward = np.sign(reward)
        return obs, reward, done, info

    def reset(self, **kwargs):
        self.total_reward = 0.0
        return self.env.reset(**kwargs)

class WarpFrame(gym.ObservationWrapper):
    def __init__(self, env, width=84, height=84, grayscale=True, dict_space_key=None):
        """
        Warp frames to 84x84 as done in the Nature paper and later work.

        If the environment uses dictionary observations, `dict_space_key` can be specified which indicates which
        observation should be warped.
        """
        super().__init__(env)
        self._width = width
        self._height = height
        self._grayscale = grayscale
        self._key = dict_space_key
        if self._grayscale:
            num_colors = 1
        else:
            num_colors = 3

        new_space = gym.spaces.Box(
            low=0,
            high=255,
            shape=(self._height, self._width, num_colors),
            dtype=np.uint8,
        )
        if self._key is None:
            original_space = self.observation_space
            self.observation_space = new_space
        else:
            original_space = self.observation_space.spaces[self._key]
            self.observation_space.spaces[self._key] = new_space
        assert original_space.dtype == np.uint8 and len(original_space.shape) == 3

    def observation(self, obs):
        if self._key is None:
            frame = obs
        else:
            frame = obs[self._key]

        if self._grayscale:
            frame = cv2.cvtColor(frame, cv2.COLOR_RGB2GRAY)
        frame = cv2.resize(
            frame, (self._width, self._height), interpolation=cv2.INTER_AREA
        )
        if self._grayscale:
            frame = np.expand_dims(frame, -1)

        if self._key is None:
            obs = frame
        else:
            obs = obs.copy()
            obs[self._key] = frame
        return obs

class ScaledFloatFrame(gym.ObservationWrapper):
    def __init__(self, env):
        gym.ObservationWrapper.__init__(self, env)
        self.observation_space = gym.spaces.Box(low=0, high=1, shape=env.observation_space.shape, dtype=np.float32)

    def observation(self, observation):
        # careful! This undoes the memory optimization, use
        # with smaller replay buffers only.
        return np.array(observation).astype(np.float32) / 255.0

class LazyFrames(object):
    def __init__(self, frames):
        """This object ensures that common frames between the observations are only stored once.
        It exists purely to optimize memory usage which can be huge for DQN's 1M frames replay
        buffers.

        This object should only be converted to numpy array before being passed to the model.

        You'd not believe how complex the previous solution was."""
        self._frames = frames
        self._out = None

    def _force(self):
        if self._out is None:
            self._out = np.concatenate(self._frames, axis=-1)
            self._frames = None
        return self._out

    def __array__(self, dtype=None):
        out = self._force()
        if dtype is not None:
            out = out.astype(dtype)
        return out

    def __len__(self):
        return len(self._force())

    def __getitem__(self, i):
        return self._force()[i]

    def count(self):
        frames = self._force()
        return frames.shape[frames.ndim - 1]

    def frame(self, i):
        return self._force()[..., i]

def wrap_deepmind(env, episode_life=True, clip_rewards=True, frame_stack=1, scale=False):
    """Configure environment for DeepMind-style Atari"""
    if episode_life:
        env = EpisodicLifeEnv(env)
    if 'FIRE' in env.unwrapped.get_action_meanings():
        env = FireResetEnv(env)
    env = WarpFrame(env)
    if scale:
        env = ScaledFloatFrame(env)
    if clip_rewards:
        env = ClipRewardEnv(env)
    env = FrameStack(env, frame_stack)
    return env




import torch
import torch.nn as nn

from .utils import Scale, Unscale
from .distributions import get_dist
from .neural_networks import NNBase, get_model


class OnPolicyActorCritic(nn.Module):
    """
    Actor critic class for On-Policy algorithms.

    It contains a policy network (actor) to predict next actions and a critic
    value network to predict the value score of a given obs.

    Parameters
    ----------
    input_space : gym.Space
        Environment observation space.
    action_space : gym.Space
        Environment action space.
    feature_extractor_network : nn.Module
        PyTorch nn.Module used as the features extraction block in all networks.
    feature_extractor_kwargs : dict
        Keyword arguments for the feature extractor network.
    recurrent_policy : bool
        Whether to use a RNN as a policy.
    recurrent_hidden_size : int
        Policy RNN hidden size.
    shared_policy_value_network : bool
        Whether or not to share weights between policy and value networks.

    Attributes
    ----------
    policy_net : nn.module
        Neural network that extracts features from the input env obs.
    input_space : gym.Space
        Environment observation space.
    action_space : gym.Space
        Environment action space.
    dist : nn.module
        Neural network that predict a prob distribution over the action space
        from the features extracted by the policy_net.
    value_net : nn.module
        Neural network that predicts a value score for a given env obs.
    shared_policy_value_network : bool
        If True, the feature extraction block of the value_net will be the
        policy_net.
    scale : nn.module
        Maps actions from [space.low, space.high] range to [-1, 1] range.
    unscale : nn.module
        Maps actions from [-1, 1] range to [space.low, space.high] range.

    """
    def __init__(self,
                 input_space,
                 action_space,
                 feature_extractor_network=get_model("MLP"),
                 feature_extractor_kwargs={},
                 recurrent_policy=False,
                 recurrent_hidden_size=512,
                 shared_policy_value_network=True):

        super(OnPolicyActorCritic, self).__init__()
        self.input_space = input_space
        self.action_space = action_space
        self.shared_policy_value_network = shared_policy_value_network

        # ---- Input/Output spaces --------------------------------------------

        policy_inputs = [input_space.shape]
        value_inputs = [input_space.shape]

        # ---- Networks -------------------------------------------------------

        self.policy_net = NNBase(
            policy_inputs, feature_extractor_network,
            feature_extractor_kwargs, recurrent=recurrent_policy,
            recurrent_hidden_size=recurrent_hidden_size, final_activation=True)

        if self.shared_policy_value_network:
            self.value_net = nn.Linear(self.policy_net.num_outputs, 1)
            self.last_actor_features = None

        else:
            self.value_net = NNBase(
                value_inputs, feature_extractor_network,
                feature_extractor_kwargs, output_shape=(1,))

        # ---- Distributions --------------------------------------------------

        if action_space.__class__.__name__ == "Discrete":
            self.dist = get_dist("Categorical")(self.policy_net.num_outputs, action_space.n)
            self.scale = None
            self.unscale = None

        elif action_space.__class__.__name__ == "Box":  # Continuous action space
            self.dist = get_dist("Gaussian")(self.policy_net.num_outputs, action_space.shape[0])
            self.scale = Scale(action_space)
            self.unscale = Unscale(action_space)

        else:
            raise NotImplementedError

    @classmethod
    def actor_critic_factory(
            cls,
            input_space,
            action_space,
            restart_model=None,
            recurrent_policy=False,
            recurrent_hidden_size=512,
            feature_extractor_kwargs={},
            shared_policy_value_network=False,
            feature_extractor_network=get_model("MLP")):
        """
        Returns a function that creates actor critic instances.

        Parameters
        ----------
        input_space : gym.Space
            Environment observation space.
        action_space : gym.Space
            Environment action space.
        restart_model : str
            Path to a previously trained ActorCritic checkpoint to be loaded.
        feature_extractor_network : nn.Module
            PyTorch nn.Module used as the features extraction block in all networks.
        feature_extractor_kwargs : dict
            Keyword arguments for the feature extractor network.
        recurrent_policy : bool
            Whether to use a RNN as a policy.
        recurrent_hidden_size : int
            Policy RNN hidden size.
        shared_policy_value_network : bool
            Whether or not to share weights between policy and value networks.

        Returns
        -------
        create_actor_critic_instance : func
            creates a new OnPolicyActorCritic class instance.
        """

        def create_actor_critic_instance(device):
            """Create and return an actor critic instance."""
            policy = cls(input_space=input_space,
                         action_space=action_space,
                         recurrent_policy=recurrent_policy,
                         recurrent_hidden_size=recurrent_hidden_size,
                         feature_extractor_kwargs=feature_extractor_kwargs,
                         feature_extractor_network=feature_extractor_network,
                         shared_policy_value_network=shared_policy_value_network)
            if restart_model:
                policy.load_state_dict(
                    torch.load(restart_model, map_location=device))
            return policy

        return create_actor_critic_instance

    @property
    def is_recurrent(self):
        """Returns True if the policy network has recurrency."""
        return self.policy_net.is_recurrent

    @property
    def recurrent_hidden_state_size(self):
        """Size of policy recurrent hidden state"""
        return self.policy_net.recurrent_hidden_state_size

    def policy_initial_states(self, obs):
        """
        Returns all policy inputs to predict the environment initial action.

        Parameters
        ----------
        obs : torch.tensor
            Initial environment observation.

        Returns
        -------
        obs : torch.tensor
            Initial environment observation.
        rhs : torch.tensor
            Initial recurrent hidden state (will contain zeroes).
        done : torch.tensor
            Initial done tensor, indicating the environment is not done.
        """
        obs, rhs, done = self.policy_net.initial_states(obs)
        return obs, rhs, done

    def get_action(self, obs, rhs, done, deterministic=False):
        """
        Predict and return next action, along with other information.

        Parameters
        ----------
        obs : torch.tensor
            Current environment observation.
        rhs : torch.tensor
            Current recurrent hidden state.
        done : torch.tensor
            Current done tensor, indicating if episode has finished.
        deterministic : bool
            Whether to randomly sample action from predicted distribution or take the mode.

        Returns
        -------
        action : torch.tensor
            Next action sampled.
        clipped_action : torch.tensor
            Next action sampled, but clipped to be within the env action space.
        logp_action : torch.tensor
            Log probability of `action` within the predicted action distribution.
        rhs : torch.tensor
            Updated recurrent hidden state.
        entropy_dist : torch.tensor
            Entropy of the predicted action distribution.
        """

        actor_features, rhs = self.policy_net(obs, rhs, done)
        self.last_actor_features = actor_features

        (action, clipped_action, logp_action, entropy_dist) = self.dist(
            actor_features, deterministic=deterministic)

        if self.unscale:
            action = self.unscale(action)
            clipped_action = self.unscale(clipped_action)

        return action, clipped_action, logp_action, rhs, entropy_dist

    def evaluate_actions(self, obs, rhs, done, action):
        """
        Evaluate log likelihood of action given obs and the current
        policy network. Returns also value prediction and entropy distribution.

        Parameters
        ----------
        obs : torch.tensor
            Environment observation.
        rhs : torch.tensor
            Recurrent hidden state.
        done : torch.tensor
            Done tensor, indicating if episode has finished.
        action : torch.tensor
            Evaluated action.

        Returns
        -------
        value : torch.tensor
            Value score according to current value_net version.
        logp_action : torch.tensor
            Log probability of `action` according to the action distribution
            predicted with current version of the policy_net.
        entropy_dist : torch.tensor
            Entropy of the action distribution predicted with current version
            of the policy_net.
        rhs : torch.tensor
            Updated recurrent hidden state.
        """

        if self.scale:
            action = self.scale(action)

        actor_features, rhs = self.policy_net(obs, rhs, done)
        logp_action, entropy_dist = self.dist.evaluate_actions(actor_features, action)

        if self.shared_policy_value_network:
            value = self.value_net(actor_features)
        else:
            value, _ = self.value_net(obs)

        return value, logp_action, entropy_dist, rhs

    def get_value(self, obs):
        """
        Return value scores of given observation.

        Parameters
        ----------
        obs : torch.tensor
            Environment observation.

        Returns
        -------
        value : torch.tensor
            value score according to current value_net version.
        """
        if self.shared_policy_value_network:
            return self.value_net(self.last_actor_features)
        else:
            val, _ = self.value_net(obs)
            return val

class OffPolicyActorCritic(nn.Module):
    """
    Actor critic class for Off-Policy algorithms.

    It contains a policy network (actor) to predict next actions and one or two
    Q networks.

    Parameters
    ----------
    input_space : gym.Space
        Environment observation space.
    action_space : gym.Space
        Environment action space.
    feature_extractor_network : nn.Module
        PyTorch nn.Module to extract features in all networks.
    feature_extractor_kwargs : dict
        Keyword arguments for the feature extractor network.
    recurrent_policy : bool
        Whether to use a RNN as a policy.
    recurrent_hidden_size : int
        Policy RNN hidden size.
    create_double_q_critic : bool
        Whether to instantiate a second Q network or not.

    Attributes
    ----------
    policy_net : nn.module
        Neural network that extracts features from the input env obs.
    input_space : gym.Space
        Environment observation space.
    action_space : gym.Space
        Environment action space.
    dist : nn.module
        Neural network that predict a prob distribution over the action space
    q1 : nn.module
        Neural network that predicts a Q-value for a given env obs and action.
    q2 : nn.module
        A second neural network to predict a Q-value for a given env obs and action.
    scale : nn.module
        Maps actions from [space.low, space.high] range to [-1, 1] range.
    unscale : nn.module
        Maps actions from [-1, 1] range to [space.low, space.high] range.

    Examples
    --------
    """
    def __init__(self,
                 input_space,
                 action_space,
                 feature_extractor_network=get_model("MLP"),
                 feature_extractor_kwargs={},
                 recurrent_policy=False,
                 recurrent_hidden_size=512,
                 create_double_q_critic=True):

        super(OffPolicyActorCritic, self).__init__()
        self.input_space = input_space
        self.action_space = action_space

        # ---- Input/Output spaces --------------------------------------------

        policy_inputs = [input_space.shape]

        if action_space.__class__.__name__ == "Discrete":
            q_inputs = [input_space.shape]
            q_outputs = (action_space.n,)

        elif action_space.__class__.__name__ == "Box":
            q_outputs = (1,)
            if len(input_space.shape) == 3: # observations are images
                q_inputs = [input_space.shape, action_space.shape]
            elif len(input_space.shape) == 1:
                q_inputs = [(input_space.shape[0] + action_space.shape[0],)]
            else:
                raise NotImplementedError

        else:
            raise NotImplementedError

        # ---- Networks -------------------------------------------------------

        self.policy_net = NNBase(
            policy_inputs, feature_extractor_network,
            feature_extractor_kwargs, recurrent=recurrent_policy,
            recurrent_hidden_size=recurrent_hidden_size, final_activation=True)

        self.q1 = NNBase(
            q_inputs, feature_extractor_network,
            feature_extractor_kwargs, output_shape=q_outputs)

        if create_double_q_critic:
            self.q2 = NNBase(
                q_inputs, feature_extractor_network,
                feature_extractor_kwargs, output_shape=q_outputs)
        else:
            self.q2 = None

        # ---- Distributions --------------------------------------------------

        if action_space.__class__.__name__ == "Discrete":
            self.dist = get_dist("Categorical")(self.policy_net.num_outputs, action_space.n)
            self.scale = None
            self.unscale = None

        elif action_space.__class__.__name__ == "Box":  # Continuous action space
            self.dist = get_dist("SquashedGaussian")(self.policy_net.num_outputs, action_space.shape[0])
            self.scale = Scale(action_space)
            self.unscale = Unscale(action_space)

        else:
            raise NotImplementedError

    @classmethod
    def actor_critic_factory(
            cls,
            input_space,
            action_space,
            restart_model=None,
            recurrent_policy=False,
            recurrent_hidden_size=512,
            feature_extractor_kwargs={},
            feature_extractor_network=get_model("MLP"),
            create_double_q_critic=True):
        """
        Returns a function that creates actor critic instances.

        Parameters
        ----------
        input_space : gym.Space
            Environment observation space.
        action_space : gym.Space
            Environment action space.
        restart_model : str
            Path to a previously trained ActorCritic checkpoint to be loaded.
        feature_extractor_network : nn.Module
            PyTorch nn.Module used as the features extraction block in all networks.
        feature_extractor_kwargs : dict
            Keyword arguments for the feature extractor network.
        recurrent_policy : bool
            Whether to use a RNN as a policy.
        recurrent_hidden_size : int
            Policy RNN hidden size.
        create_double_q_critic : bool
            whether to instantiate a second Q network or not.

        Returns
        -------
        create_actor_critic_instance : func
            creates a new OffPolicyActorCritic class instance.
        """

        def create_actor_critic_instance(device):
            """Create and return an actor critic instance."""
            policy = cls(input_space=input_space,
                         action_space=action_space,
                         recurrent_policy=recurrent_policy,
                         recurrent_hidden_size=recurrent_hidden_size,
                         create_double_q_critic=create_double_q_critic,
                         feature_extractor_kwargs=feature_extractor_kwargs,
                         feature_extractor_network=feature_extractor_network)
            if restart_model:
                policy.load_state_dict(
                    torch.load(restart_model, map_location=device))
            return policy

        return create_actor_critic_instance

    @property
    def is_recurrent(self):
        """Returns True if the policy network has recurrency."""
        return self.policy_net.is_recurrent

    @property
    def recurrent_hidden_state_size(self):
        """Size of policy recurrent hidden state"""
        return self.policy_net.recurrent_hidden_state_size

    def policy_initial_states(self, obs):
        """
        Returns all policy inputs to predict the environment initial action.

        Parameters
        ----------
        obs : torch.tensor
            Initial environment observation.

        Returns
        -------
        obs : torch.tensor
            Initial environment observation.
        rhs : torch.tensor
            Initial recurrent hidden state (will contain zeroes).
        done : torch.tensor
            Initial done tensor, indicating the environment is not done.
        """
        obs, rhs, done = self.policy_net.initial_states(obs)
        return obs, rhs, done

    def get_action(self, obs, rhs, dones, deterministic=False):
        """
        Predict and return next action, along with other information.

        Parameters
        ----------
        obs : torch.tensor
            Current environment observation.
        rhs : torch.tensor
            Current recurrent hidden state.
        done : torch.tensor
            Current done tensor, indicating if episode has finished.
        deterministic : bool
            Whether to randomly sample action from predicted distribution or take the mode.

        Returns
        -------
        action : torch.tensor
            Next action sampled.
        clipped_action : torch.tensor
            Next action sampled, but clipped to be within the env action space.
        logp_action : torch.tensor
            Log probability of `action` within the predicted action distribution.
        rhs : torch.tensor
            Updated recurrent hidden state.
        entropy_dist : torch.tensor
            Entropy of the predicted action distribution.
        """
        actor_features, rhs = self.policy_net(obs, rhs, dones)
        self.last_actor_features = actor_features

        (action, clipped_action, logp_action, entropy_dist) = self.dist(
            actor_features, deterministic=deterministic)

        if self.unscale:
            action = self.unscale(action)
            clipped_action = self.unscale(clipped_action)

        return action, clipped_action, logp_action, rhs, entropy_dist

    def get_q_scores(self, obs, actions=None):
        """
        Return Q scores of the given observations and actions.
        Parameters
        ----------
        obs : torch.tensor
            Environment observation.
        actions : torch.tensor
             Evaluated actions.

        Returns
        -------
        q1 : torch.tensor
            Q score according to current q1 network version.
        q2 : torch.tensor
            Q score according to current q2 network version.
        """

        if actions is not None:
            if len(obs.shape[1:]) == 3: # obs are images
                inputs = (obs, actions)
            elif len(obs.shape[1:]) == 1:
                inputs = torch.cat([obs, actions], dim=-1)
        else:
            inputs = obs

        q1, _ = self.q1(inputs)
        q2, _ = self.q2(inputs) if self.q2 else None
        return q1, q2


import os
import obstacle_tower_env
from obstacle_tower_env import ObstacleTowerEnv
from ..common import FrameStack, FrameSkip
from .wrappers import (ReducedActionEnv, BasicObstacleEnv,
                       RewardShapeObstacleEnv, BasicObstacleEnvTest)

# info_keywords=('floor', 'start', 'seed'),

def make_obstacle_train_env(index_worker=0, index_env=0, realtime=False, seed=0, frame_skip=0, frame_stack=1):
    """
    Create train Obstacle Tower Unity3D challenge environment.

    Parameters
    ----------
    index_worker : int
        Index of the worker running this environment.
    index_env : int
        Index of this environment withing the vector of environments.
    seed : int
        Used as base worker id.
    frame_skip : int
        Return only every `frame_skip`-th observation.
    frame_stack : int
        Observations composed of last `frame_stack` frames stacked.
    realtime : bool
        Visualise environment.

    Returns
    -------
    env : gym.Env
        Train environment.
    """
    if 'DISPLAY' not in os.environ.keys():
        os.environ['DISPLAY'] = ':0'

    exe = os.path.join(
        os.path.dirname(obstacle_tower_env.__file__),
        'ObstacleTower/obstacletower')

    id = seed + index_worker + index_env
    env = ObstacleTowerEnv(
        environment_filename=exe, retro=True, worker_id=id, greyscale=False,
        docker_training=False, realtime_mode=realtime)

    env = ReducedActionEnv(env)
    env = BasicObstacleEnv(env, max_floor=50, min_floor=0)
    env = RewardShapeObstacleEnv(env)

    if frame_skip > 0:
        env = FrameSkip(env, skip=frame_skip)

    if frame_stack > 1:
        env = FrameStack(env, k=frame_stack)

    return env

def make_obstacle_test_env(index_worker=0, index_env=0, realtime=False, seed=0, frame_skip=0, frame_stack=1):
    """
    Create test Obstacle Tower Unity3D challenge environment.

    Parameters
    ----------
    index_worker : int
        Index of the worker running this environment.
    index_env : int
        Index of this environment withing the vector of environments.
    seed : int
        Used as base worker id.
    frame_skip : int
        Return only every `frame_skip`-th observation.
    frame_stack : int
        Observations composed of last `frame_stack` frames stacked.
    realtime : bool
        Visualise environment.

    Returns
    -------
    env : gym.Env
        Test environment.
    """
    if 'DISPLAY' not in os.environ.keys():
        os.environ['DISPLAY'] = ':0'

    exe = os.path.join(os.path.dirname(obstacle_tower_env.__file__),
                       'ObstacleTower/obstacletower')
    id = seed + index_worker + index_env
    env = ObstacleTowerEnv(
        environment_filename=exe, retro=True, worker_id=id, greyscale=False,
        docker_training=False, realtime_mode=realtime)

    env = ReducedActionEnv(env)
    env = BasicObstacleEnvTest(env, max_floor=50, min_floor=0)

    if frame_skip > 0:
        env = FrameSkip(env, skip=frame_skip)

    if frame_stack > 1:
        env = FrameStack(env, k=frame_stack)

    return env



import os
import json
import random
import numpy as np

import gym
from gym import spaces
from nappo.envs import (
    box_is_placed, box_location, place_location,
    reduced_action_lookup_6, reduced_action_lookup_7)


class BasicObstacleEnv(gym.Wrapper):
    def __init__(self, env, min_floor, max_floor):

        gym.Wrapper.__init__(self, env)

        self.reached_floor = 0
        self._min_floor = min_floor
        self._max_floor = max_floor
        self.count = 0
        self.last_time = 3000

        self.seed = None
        self.start_floor = None

    def step(self, action):
        obs, reward, done, info = self.env.step(action)

        if info['current_floor'] > self.reached_floor:
            self.reached_floor = info['current_floor']

        if info['current_floor'] > self._max_floor:
            done = True

        info['seed'] = self.seed
        info['start'] = self.start_floor
        info['floor'] = self.reached_floor

        num_keys = info["total_keys"]
        self.picked_key = num_keys > self._previous_keys
        self._previous_keys = num_keys
        self.last_time = info['time_remaining']

        return obs, reward, done, info

    def reset(self, **kwargs):

        self._previous_keys = 0
        self.puzzle_solved = False

        self.seed = np.random.randint(0, 100)
        self.env.unwrapped.seed(self.seed)

        self.start_floor = np.random.randint(
            self._min_floor, self.reached_floor if self.reached_floor != 0 else 1)
        self.env.unwrapped.floor(self.start_floor)
        self.reached_floor = 0

        config = {"total-floors": self._max_floor + 2}
        self.count += 1

        return self.env.reset(config=config, **kwargs)

class BasicObstacleEnvTest(gym.Wrapper):
    def __init__(self, env, min_floor, max_floor, seed_list=[1001, 1002, 1003, 1004, 1005]):

        gym.Wrapper.__init__(self, env)

        self.reached_floor = None
        self._min_floor = min_floor
        self._max_floor = max_floor
        self.last_time = 3000

        self.seed_index = -1
        self.seed_list = seed_list
        self.reached_floors = []

    def step(self, action):
        obs, reward, done, info = self.env.step(action)

        if info['current_floor'] > self.reached_floor:
            self.reached_floor = info['current_floor']

        info['seed'] = self.seed
        info['start'] = self.start_floor
        info['floor'] = self.reached_floor

        num_keys = info["total_keys"]
        self.picked_key = num_keys > self._previous_keys
        self._previous_keys = num_keys
        self.last_time = info['time_remaining']

        return obs, reward, done, info

    def reset(self, **kwargs):

        if self.reached_floor is not None:
            self.reached_floors.append(self.reached_floor)
            print("Seed {}, reahed foor {}".format(self.seed, self.reached_floor))
        if self.seed_index == len(self.seed_list) - 1:
            print("Average of all seeds {}".format(np.mean(self.reached_floors)))
            self.reached_floors = []

        self._previous_keys = 0
        self.seed_index = (self.seed_index + 1) % len(self.seed_list)
        self.seed = self.seed_list[self.seed_index]
        self.env.unwrapped.seed(self.seed)

        self.start_floor = self._min_floor
        self.env.unwrapped.floor(self.start_floor)
        self.reached_floor = 0
        config = {"total-floors": self._max_floor + 2}

        return self.env.reset(config=config, **kwargs)

class RewardShapeObstacleEnv(gym.Wrapper):

    def __init__(self, env, killed_reward=2):
        gym.Wrapper.__init__(self, env)
        self._killed_reward = killed_reward
        self.time_remaining = None

    def step(self, action):
        obs, reward, done, info = self.env.step(action)

        # Picked key
        if self.env.picked_key:
            reward += 1.01

        found_box, _ = box_location(obs)
        if found_box:
            reward += 0.002

        found_place, _ = place_location(obs)
        if found_place:
            reward += 0.001

        # Account for solving puzzle
        if box_is_placed(obs) and (not self.puzzle_solved) and (reward > 0.08):
            self.puzzle_solved = True
            reward += 1.5

        if self.time_remaining == None:
            self.time_remaining = info['time_remaining']
        if info['time_remaining'] > self.time_remaining:
            reward += 0.002

        return obs, reward, done, info

    def reset(self, **kwargs):
        self.puzzle_solved = False
        self.time_remaining = None
        return self.env.reset(**kwargs)

class ReducedActionEnv(gym.Wrapper):
    def __init__(self, env, num_actions=6):

        if num_actions == 6:
            _action_lookup = reduced_action_lookup_6
        elif num_actions == 7:
            _action_lookup = reduced_action_lookup_7
        else:
            ValueError("No lookup table for num reduced actions {}".format(
                num_actions))

        env.unwrapped._flattener.action_lookup = _action_lookup
        num_actions = len(env.unwrapped._flattener.action_lookup)
        env.unwrapped._flattener.action_space = spaces.Discrete(num_actions)
        env.unwrapped._action_space = env.unwrapped._flattener.action_space
        gym.Wrapper.__init__(self, env)




import ray
from .g_worker import GWorker
from ..base.worker_set import WorkerSet as WS
from ..base.worker import default_remote_config

class GWorkerSet(WS):
    """
    Class to better handle the operations of ensembles of workers.

    Parameters
    ----------
    create_algo_instance : func
        A function that creates an algorithm class.
    create_storage_instance : func
        A function that create a rollouts storage.
    create_actor_critic_instance : func
        A function that creates a policy.
    create_collection_worker_set_instance : func
        A function that creates worker sets of data collection workers.
    worker_remote_config : dict
        Ray resource specs for the remote workers.
    max_collector_workers_requests_pending : int
        maximum number of collection tasks to simultaneously scheduled to
        collection workers.
    num_workers : int
        Number of remote workers in the worker set.

    Attributes
    ----------
    worker_class : python class
        Worker class to be instantiated to create Ray remote actors.
    remote_config : dict
        Ray resource specs for the remote workers.
    worker_params : dict
        Keyword parameters of the worker_class.
    num_workers : int
        Number of remote workers in the worker set.
    """

    def __init__(self,
                 create_algo_instance,
                 create_storage_instance,
                 create_actor_critic_instance,
                 create_collection_worker_set_instance,
                 worker_remote_config=default_remote_config,
                 max_collector_workers_requests_pending=2,
                 num_workers=1):
        """
        Initialize a WorkerSet.

        Arguments:
            create_algo_instance (func): a function that creates an algorithm class.
            create_storage_instance (func): a function that create a rollouts storage.
            create_actor_critic_instance (func): a function that creates a policy.
            max_collector_workers_requests_pending (int): maximum number of collect() tasks to schedule for each collection worker.
            worker_remote_config (dict): ray resources specs for the gradient workers.
            num_workers (int): number of gradient computing workers.
        """

        self.worker_class = GWorker
        default_remote_config.update(worker_remote_config)
        self.remote_config = default_remote_config
        self.worker_params = {
            "create_algo_instance": create_algo_instance,
            "create_storage_instance": create_storage_instance,
            "create_actor_critic_instance": create_actor_critic_instance,
            "max_collect_requests_pending": max_collector_workers_requests_pending,
            "create_collection_worker_set_instance": create_collection_worker_set_instance,
        }

        super(GWorkerSet, self).__init__(
            worker=self.worker_class,
            worker_params=self.worker_params,
            worker_remote_config=self.remote_config,
            num_workers=num_workers)


import torch


@torch.no_grad()
def get_updated_action_log_probs(storage, new_ac, algo):
    """
    Arguments:
        storage ():
        new_policy (ActorCritic):
        algo ():
    """

    len, num_proc = storage.data["act"].shape[0:2]

    # Create batches without shuffling data
    batches = storage.generate_batches(
        algo.num_mini_batch, algo.mini_batch_size,
        num_epochs=1, recurrent_ac=new_ac.is_recurrent, shuffle=False)

    # Obtain new value and log probability predictions
    new_val = []
    new_logp = []
    for batch in batches:
        obs, rhs, act, done = batch["obs"], batch["rhs"], batch["act"], batch["done"]
        (val, logp, _, _) = new_ac.evaluate_actions(obs, rhs, done, act)
        new_val.append(val)
        new_logp.append(logp)

    # Concatenate results
    if new_ac.is_recurrent:
        new_val = [p.view(len, num_proc // algo.num_mini_batch, -1) for p in new_val]
        storage.data["val"][:-1] = torch.cat(new_val, dim=1)
        new_logp = [p.view(len, num_proc // algo.num_mini_batch, -1) for p in new_logp]
        new_logp = torch.cat(new_logp, dim=1)
    else:
        storage.data["val"][:-1] = torch.cat(new_val, dim=0).view(len, num_proc, 1)
        new_logp = torch.cat(new_logp, dim=0).view(len, num_proc, 1)

    return new_logp

@torch.no_grad()
def compute_vtrace(storage, new_policy, algo, clip_rho_thres=1.0, clip_c_thres=1.0):
    """
    Arguments:
        storage ():
        new_policy (Actorritic):
        algo ():
        clip_rho_thres (float):
        clip_c_thres (float):
    """

    l = storage.step if storage.step != 0 else storage.max_size

    with torch.no_grad():
        _ = new_policy.get_action(
            storage.data["obs"][storage.step - 1],
            storage.data["rhs"][storage.step - 1],
            storage.data["done"][storage.step - 1])
        next_value = new_policy.get_value(storage.data["obs"][storage.step - 1])
    storage.data["val"][storage.step] = next_value

    new_action_log_probs = get_updated_action_log_probs(storage, new_policy, algo)

    log_rhos = (new_action_log_probs - storage.data["logp"])
    clipped_rhos = torch.clamp(torch.exp(log_rhos), max=clip_rho_thres)
    clipped_cs = torch.clamp(torch.exp(log_rhos), max=clip_c_thres)

    deltas = clipped_rhos * (storage.data["rew"] + algo.gamma * storage.data["val"][1:]
                             - storage.data["val"][:-1])

    acc = torch.zeros_like(storage.data["val"][-1])
    result = []
    for i in reversed(range(l)):
        acc = deltas[i] + algo.gamma * clipped_cs[i] * acc * (1 - storage.data["done"][i + 1])
        result.append(acc)

    result.reverse()
    result.append(torch.zeros_like(storage.data["val"][-1]))
    vs_minus_v_xs = torch.stack(result)

    vs = torch.add(vs_minus_v_xs, storage.data["val"])
    adv = clipped_rhos * (storage.data["rew"] + algo.gamma * vs[1:] - storage.data["val"][:-1])

    storage.data["ret"] = vs
    storage.data["logp"] = new_action_log_probs
    storage.data["adv"] = (adv - adv.mean()) / (adv.std() + 1e-8)

    return storage





import ray
import socket
from contextlib import closing

def broadcast_message(key, message):
    """ _"""
    ray.worker.global_worker.redis_client.set(key, message)

def check_message(key):
    """ _"""
    return ray.worker.global_worker.redis_client.get(key)

# -----------------------------------------------------------------------------

def find_free_port():
    """
    from https://github.com/ray-project/ray/blob/master/python/ray/util/sgd/utils.py
    """
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


import os
import shutil

def cleanup_log_dir(log_dir):
    """
    Create log directory and remove old files.

    Parameters
    ----------
    log_dir : str
        Path to log directory.
    """
    try:
        os.makedirs(log_dir)
    except OSError:
        if os.path.isdir(os.path.join(log_dir, "train")):
            shutil.rmtree(os.path.join(log_dir, "train"))
        if os.path.isdir(os.path.join(log_dir, "test")):
            shutil.rmtree(os.path.join(log_dir, "test"))
        if os.path.isdir(os.path.join(log_dir, "tensorboard_logs")):
            shutil.rmtree(os.path.join(log_dir, "tensorboard_logs"))




from .replay_buffer import ReplayBuffer
from .on_policy_buffer import OnPolicyBuffer
from .on_policy_gae_buffer import OnPolicyGAEBuffer
from .hindsight_experience_replay_buffer import HindsightExperienceReplayBuffer



from .atari.make_atari_env import make_atari_train_env, make_atari_test_env
from .mujoco.make_mujoco_env import make_mujoco_train_env, make_mujoco_test_env
from .pybullet.make_pybullet_env import make_pybullet_train_env, make_pybullet_test_env
from .obstacle_tower.make_obstacle_env import make_obstacle_train_env, make_obstacle_test_env



import ray
from .worker import default_remote_config

class WorkerSet:
    """
    Class to better handle the operations of ensembles of Workers.
    Contains common functionality across all worker sets.

    Parameters
    ----------
    create_algo_instance : func
        A function that creates an algorithm class.
    create_storage_instance : func
        A function that create a rollouts storage.
    create_train_envs_instance : func
        A function to create train environments.
    create_actor_critic_instance : func
        A function that creates a policy.
    create_test_envs_instance : func
        A function to create test environments.
    worker_remote_config : dict
        Ray resource specs for the remote workers.
    num_workers : int
        Number of remote workers in the worker set.

    Attributes
    ----------
    worker_class : python class
        Worker class to be instantiated to create Ray remote actors.
    remote_config : dict
        Ray resource specs for the remote workers.
    worker_params : dict
        Keyword arguments of the worker_class.
    num_workers : int
        Number of remote workers in the worker set.
    """

    def __init__(self,
                 worker,
                 worker_params,
                 worker_remote_config=default_remote_config,
                 num_workers=1,
                 add_local_worker=True):

        self.worker_class = worker
        self.worker_params = worker_params
        self.remote_config = worker_remote_config

        if add_local_worker:
            self._local_worker = self._make_worker(
                self.worker_class, index_worker=1,
                worker_params=self.worker_params)
            self.worker_params.update({
                "initial_weights": ray.put({
                    "update": 0,
                    "weights": self._local_worker.get_weights()})})

        else:
            self._local_worker = None

        self._remote_workers = []
        self.add_workers(num_workers)
        self.num_workers = num_workers

    @staticmethod
    def _make_worker(cls, index_worker, worker_params):
        """
        Create a single worker.

        Parameters
        ----------
        index_worker : int
            Index assigned to remote worker.
        worker_params : dict
            Keyword parameters of the worker_class.

        Returns
        -------
        w : python class
            An instance of worker class cls
        """
        w = cls(index_worker=index_worker, **worker_params)
        return w

    def add_workers(self, num_workers):
        """
        Create and add a number of remote workers to this worker set.

        Parameters
        ----------
        num_workers : int
            Number of remote workers to create.
        """
        cls = self.worker_class.as_remote(**self.remote_config).remote
        self._remote_workers.extend([
            self._make_worker(cls, index_worker=i + 1, worker_params=self.worker_params)
            for i in range(num_workers)])

    def local_worker(self):
        """Return local worker"""
        return self._local_worker

    def remote_workers(self):
        """Returns list of remote workers"""
        return self._remote_workers

    def stop(self):
        """Stop all remote workers"""
        for w in self.remote_workers():
            w.__ray_terminate__.remote()

